/* libdeps.c - Retrieve and calculate dependencies for deborphan.
   Copyright (C) 2000, 2001, 2002, 2003 Cris van Pelt
   Copyright (C) 2003, 2004 Peter Palfrader

   $Id: libdeps.c 409 2004-04-21 12:20:43Z weasel $

   Distributed under the terms of the Artistic License.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <regex.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <deborphan.h>

#ifdef USE_XALLOC
#  include <xalloc.h>
#endif

extern int options[];

static void
deactivate_deps(pkg_info *all_packages, pkg_info *current_pkg);
static pkg_info**
find_uphill_deps(pkg_info** uphill_deps,
                 int *uphill_cnt, int *uphill_max,
                 pkg_info *all_packages, pkg_info *current_pkg);
static pkg_info**
find_downhill_deps(pkg_info** downhill_deps,
                   int *downhill_cnt, int *downhill_max,
                   pkg_info *all_packages, pkg_info *current_pkg);

/* Returns 1 iif child depends on father */
static int
depends_on(pkg_info *child, pkg_info *father)
{
    int dep, prov;

    for (dep = 0; dep < child->deps_cnt; dep++) {
        for (prov = 0; prov < father->provides_cnt; prov++) {
            if (pkgcmp(father->provides[prov], child->deps[dep]))
                return 1;
        }
        if (pkgcmp(father->self, child->deps[dep]))
            return 1;
    }
    return 0;
}

/* For each package found, this scans the `package' structure, to
 * see if anything depends on it.
 */
void
check_lib_deps(pkg_info * all_packages, pkg_info * current_pkg)
{
    int deps, prov, no_dep_found = 1, search_found = 1;
    int i;
    pkg_info *package;
    static int j;

    extern char **search_for;

    if (options[FIND_CONFIG] && !current_pkg->config)
	return;
    if (current_pkg->hold)
	return;
    if (current_pkg->priority < options[PRIORITY])
	return;
    if (keep && mustkeep(current_pkg->self))
	return;
    if (!is_library(current_pkg, options[SEARCH_LIBDEVEL]))
	return;

    if (options[SEARCH]) {
	search_found = 0;
	/* Count packages. Just once. */
	if (!j)
	    for (; *(search_for+j); j++);

	/* Search for the package, and clear it from the list if it is
	   found. */
	for (i = 0; *(search_for + i); i++) {
	    if (!strcmp(*(search_for + i), current_pkg->self.name)) {
		search_found = 1;
		*(search_for + i) = *(search_for + --j);
		*(search_for + j) = NULL;
		break;
	    }
	}
    }
    if (!search_found)
	return;

    if (options[SHOW_DEPS]) {
	printf("%s", current_pkg->self.name);

	if (options[SHOW_SECTION])
	    printf(" (%s", current_pkg->section);
	if (options[SHOW_PRIORITY])
	    printf(" - %s", priority_to_string(current_pkg->priority));
	if (options[SHOW_SIZE])
	    printf(", %ld", current_pkg->installed_size);
	if (options[SHOW_SECTION])
	    printf(")");
	printf("\n");
    }

    /* Search all (installed) packages for dependencies. 
     */
    for (package = all_packages; package && no_dep_found; package = package->next) {
	for (deps = 0; deps < package->deps_cnt && no_dep_found; deps++) {
	    for (prov = 0; prov < current_pkg->provides_cnt && no_dep_found;
		 prov++) {
		if (pkgcmp(current_pkg->provides[prov], package->deps[deps])) {
		    if (options[SHOW_DEPS])
			printf("      %s\n", package->self.name);
		    else
			no_dep_found = 0;
		}
	    }

	    if (pkgcmp(current_pkg->self, package->deps[deps])) {
		if (options[SHOW_DEPS])
		    printf("      %s\n", package->self.name);
		else
		    no_dep_found = 0;
	    }
	}
    }

    if (no_dep_found && (!options[SHOW_DEPS])) {
	if (options[SHOW_SIZE])
	    printf("%10ld ", current_pkg->installed_size);

	if (options[SHOW_SECTION])
	    printf("%-25s", current_pkg->section);

	if (options[SHOW_PRIORITY])
	    printf("%-24s %s", current_pkg->self.name,
		   priority_to_string(current_pkg->priority));
	else
	    printf("%s", current_pkg->self.name);

	printf("\n");

        if (options[FIND_CIRCULAR])
            deactivate_deps(all_packages, current_pkg);
    }
}

static void
deactivate_deps(pkg_info *all_packages, pkg_info *current_pkg)
{
    pkg_info *pkg;

    current_pkg->activity = DEACTIVATED;
    for (pkg = all_packages; pkg; pkg = pkg->next) {
        if (pkg->activity == NORMAL && depends_on(current_pkg, pkg))
            deactivate_deps(all_packages, pkg);
    }
}

void find_remaining_circular_deps(pkg_info* all_packages)
{
    pkg_info *pkg;
    pkg_info **downhill_deps, **uphill_deps;
    int downhill_cnt, uphill_cnt;
    int downhill_max, uphill_max;
    int i, chars_written;

    downhill_max = INIT_DEPENDS_COUNT;
    downhill_cnt = 0;
    downhill_deps = xmalloc(downhill_max * sizeof(pkg_info*));
    memset(downhill_deps, '\0', downhill_max * sizeof(pkg_info*));
    uphill_max = INIT_DEPENDS_COUNT;
    uphill_cnt = 0;
    uphill_deps = xmalloc(uphill_max * sizeof(pkg_info*));
    memset(uphill_deps, '\0', uphill_max * sizeof(pkg_info*));

    for (pkg = all_packages; pkg; pkg = pkg->next) {
        if (pkg->activity == DEACTIVATED)
            continue;
        if (pkg->hold ||
            pkg->priority < options[PRIORITY] ||
            keep && mustkeep(pkg->self) ||
            !is_library(pkg, options[SEARCH_LIBDEVEL])) {
            deactivate_deps(all_packages, pkg);
            continue;
        }
        // First, find all that depend on me
        uphill_cnt = 0;
        uphill_deps = find_uphill_deps(uphill_deps, &uphill_cnt, &uphill_max,
                                       all_packages, pkg);
        // Now, all that I depend on
        downhill_cnt = 0;
        downhill_deps = find_downhill_deps(downhill_deps, &downhill_cnt, &downhill_max,
                                           all_packages, pkg);
        // Make sure I depend on all that depend on me
        for (i = 0; i < uphill_cnt; i++)
            if (uphill_deps[i]->activity == UPHILL) // Pb: "should" have been turned to DOWNHILL
                break;
        if (i == uphill_cnt) { // Yes!
            // Compute total installed size (if asked to)
            if (options[SHOW_SIZE]) {
                int total_size = 0;
                for (i = 0; i < uphill_cnt; i++)
                    total_size += uphill_deps[i]->installed_size; 
                printf("%10ld ", total_size);
            }
            // Attempt to find common section (if asked to)
            if (options[SHOW_SECTION]) {
                char *section = NULL;
                section = strdup(uphill_deps[0]->section);
                for (i = 1; i < uphill_cnt; i++)
                    if (strcmp(section, uphill_deps[i]->section) != 0) {
                        free(section);
                        section = strdup("*");
                        break;
                    }
                printf("%-25s", section);
                free(section);
            }
            // Print the names
            chars_written = 0;
            for (i = 0; i < uphill_cnt; i++) {
                if (i) {
                    printf(",");
                    chars_written += 1;
                }
                printf(uphill_deps[i]->self.name);
                chars_written += strlen(uphill_deps[i]->self.name);
            }
            // Print highest priority among the deps (if asked to)
            if (options[SHOW_PRIORITY]) {
                int highest_priority = 6; // 1 is higher than 5...
                const char *priority_str;
                for (i = 0; i < uphill_cnt; i++)
                    if (highest_priority > uphill_deps[i]->priority)
                        highest_priority = uphill_deps[i]->priority;
                priority_str = priority_to_string(highest_priority);
                printf(" %*s", (24 > chars_written ?
                                24 - chars_written + strlen(priority_str) : 0),
                       priority_str);
            }
            printf("\n");
            // Deactivate everything
            for (i = 0; i < downhill_cnt; i++)
                downhill_deps[i]->activity = DEACTIVATED;
        } else { // No: there is a smaller circular dependency
            // Just like in the other branch, we should deactivate all
            // downhill, as only those in uphill and not in downhill
            // can have a smaller circular dependency
            for (i = 0; i < downhill_cnt; i++)
                downhill_deps[i]->activity = DEACTIVATED;
            for (i = 0; i < uphill_cnt; i++)
                if (uphill_deps[i]->activity == UPHILL) // thus, not in downhill
                    uphill_deps[i]->activity = NORMAL;
        }
    }
}

static pkg_info**
find_uphill_deps(pkg_info** uphill_deps,
                 int *uphill_cnt, int *uphill_max,
                 pkg_info *all_packages, pkg_info *current_pkg)
{
    pkg_info *pkg;

    if (*uphill_cnt >= *uphill_max) {
        *uphill_max *= 2;
        uphill_deps = xrealloc(uphill_deps, *uphill_max * sizeof(pkg_info*));
    }
    uphill_deps[ (*uphill_cnt)++ ] = current_pkg;
    current_pkg->activity = UPHILL;

    for (pkg = all_packages; pkg; pkg = pkg->next) {
        if (pkg->activity == NORMAL && depends_on(pkg, current_pkg))
            uphill_deps = find_uphill_deps(uphill_deps, uphill_cnt, uphill_max,
                                           all_packages, pkg);
    }
    return uphill_deps;
}


static pkg_info**
find_downhill_deps(pkg_info** downhill_deps,
                   int *downhill_cnt, int *downhill_max,
                   pkg_info *all_packages, pkg_info *current_pkg)
{
    pkg_info *pkg;

    if (*downhill_cnt >= *downhill_max) {
        *downhill_max *= 2;
        downhill_deps = xrealloc(downhill_deps, *downhill_max * sizeof(pkg_info*));
    }
    downhill_deps[ (*downhill_cnt)++ ] = current_pkg;
    current_pkg->activity = DOWNHILL;

    for (pkg = all_packages; pkg; pkg = pkg->next) {
        if ((pkg->activity == NORMAL || pkg->activity == UPHILL) &&
            depends_on(current_pkg, pkg))
            downhill_deps = find_downhill_deps(downhill_deps, downhill_cnt,
                                               downhill_max, all_packages, pkg);
    }
    return downhill_deps;
}
